package com.longfellowchen.btlib;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothAdapter.LeScanCallback;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattService;
import android.bluetooth.BluetoothGattDescriptor;
import android.bluetooth.BluetoothManager;
import android.content.Intent;
import android.os.Build.VERSION;
import android.os.Handler;
import android.util.Log;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.HashMap;
import java.util.Map;
import java.nio.ByteBuffer;

public class BluetoothLeDriver {
    private static final String TAG = "BluetoothLeDriver";
    private static final long SCAN_PERIOD = 3000;
    private BluetoothAdapter mBluetoothAdapter;
    private BluetoothGatt mBluetoothGatt = null;
    private Handler mHandler = new Handler();

    Map<String, BluetoothGattService> mapGattServices = new HashMap();
    Map<String, BluetoothGattCharacteristic> mapGattCharacteristic = new HashMap();
    Map<String, BluetoothDevice> mapBluetoothDevice = new HashMap();
    
    private LeScanCallback mLeScanCallback = new LeScanCallback() {
        public void onLeScan(BluetoothDevice device, int rssi, byte[] scanRecord) {
            BluetoothDevice dev = mapBluetoothDevice.get(device.getAddress());
            if( dev == null )
            {
                mapBluetoothDevice.put(device.getAddress(),device);
                Log.d(TAG,"addDevice:name:" + device.getName() + ",address:" + device.getAddress());
                BtlibPlugin.me.ThrowEvent(BTEvent.BLE_SCAN_RESULT, "name", device.getName(), "address", device.getAddress(), "rssi", Integer.toString(rssi) );
            }
        }
    };
    private boolean mScanning;

    private void doStartScan() {
        mHandler.postDelayed(new Runnable() {
            public void run() {
                if (mScanning) {
                    mScanning = false;
                    BtlibPlugin.me.ThrowEvent(BTEvent.BLE_STOP_SCAN);
                    mBluetoothAdapter.stopLeScan(mLeScanCallback);
                }
            }
        }, SCAN_PERIOD);
        mScanning = true;
        mapBluetoothDevice.clear();
        BtlibPlugin.me.ThrowEvent(BTEvent.BLE_START_SCAN);
        mBluetoothAdapter.startLeScan(mLeScanCallback);
    }

    public void stopScan() {
        mScanning = false;
        if (mBluetoothAdapter != null) {
            BtlibPlugin.me.ThrowEvent(BTEvent.BLE_STOP_SCAN);
            mBluetoothAdapter.stopLeScan(mLeScanCallback);
        }
    }

    public static byte[] UUIDasBytes(UUID uuid) {
        ByteBuffer bb = ByteBuffer.wrap(new byte[16]);
        bb.putLong(uuid.getMostSignificantBits());
        bb.putLong(uuid.getLeastSignificantBits());
        return bb.array();
    }

    private String UUIDtoBTString(String uuid) {
        if( uuid.length() == 4 )
        {
            return uuid.toUpperCase();
        } else {
            try {
                byte[] uuidBytes = UUIDasBytes(UUID.fromString(uuid));
                if (uuidBytes.length == 16) {
                    return String.format("%02X%02X", uuidBytes[2], uuidBytes[3]);
                }
            } catch(IllegalArgumentException e) {
                return uuid.toUpperCase();
            }
        }
        return uuid.toUpperCase();
    }

    private void lookupGattServices(BluetoothGatt gatt, List<BluetoothGattService> gattServices) {
        if (gattServices != null) {
            for (BluetoothGattService gattService : gattServices) {
                UUID uuid_serv = gattService.getUuid();
                String uuid_serv_str = UUIDtoBTString(uuid_serv.toString());
                mapGattServices.put(uuid_serv_str, gattService);
                Log.d(TAG,"lookupGattServices:GattServices - " + uuid_serv_str);
                List<BluetoothGattCharacteristic> gattCharacteristics = gattService.getCharacteristics();
                for (BluetoothGattCharacteristic gattCharacteristic : gattCharacteristics) {
                    String uuid_char_str = UUIDtoBTString(gattCharacteristic.getUuid().toString());
                    mapGattCharacteristic.put(gatt.getDevice().getAddress()+":"+uuid_serv_str + ":" + uuid_char_str, gattCharacteristic);
                    Log.d(TAG,"Characteristic:GattCharacteristic - " + uuid_char_str);
                    BtlibPlugin.me.ThrowEvent(
                        BTEvent.BLE_SERVICES_CHARACTERISTICS, 
                        "address", gatt.getDevice().getAddress(),
                        "service", uuid_serv_str,
                        "characteristic", uuid_char_str
                        );
                }
            }
        }
    }

    public void readCharacteristic(String address, String service, String characteristic) {
        BluetoothGattCharacteristic ch = mapGattCharacteristic.get(address+":"+UUIDtoBTString(service) + ":" + UUIDtoBTString(characteristic));
        if (ch != null) {
            readCharacteristic(ch);
            Log.d(TAG,"readCharacteristic");
        }
    }

    public void writeCharacteristic(String address, String service, String characteristic, byte[] out) {
        BluetoothGattCharacteristic ch = mapGattCharacteristic.get(address+":"+UUIDtoBTString(service) + ":" + UUIDtoBTString(characteristic));
        if (ch != null) {
            ch.setValue(out);
            writeCharacteristic(ch);
            Log.d(TAG, "writeCharacteristic");
        }
    }

    public void setCharacteristicNotification(String address, String service, String characteristic, boolean enabled) {
        BluetoothGattCharacteristic ch = mapGattCharacteristic.get(address+":"+UUIDtoBTString(service) + ":" + UUIDtoBTString(characteristic));
        if (ch != null) {
            setCharacteristicNotification(ch, enabled);
            Log.d(TAG, "setCharacteristicNotification");
        }
    }

    public void startScan() {
        if (!BtlibPlugin.me.activity.getPackageManager().hasSystemFeature("android.hardware.bluetooth_le")) {
            ///Toast.makeText(BtlibPlugin.me.activity, "ble_not_supported", 0).show();
            Log.e(TAG, "ble_not_supported" );
        }
        if (VERSION.SDK_INT >= 18) {
            mBluetoothAdapter = ((BluetoothManager) BtlibPlugin.me.activity.getSystemService("bluetooth")).getAdapter();
        } else {
            mBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
        }
        if (mBluetoothAdapter == null) {
            ///Toast.makeText(BtlibPlugin.me.activity, "error_bluetooth_not_supported", 0).show();
            Log.e(TAG, "error_bluetooth_not_supported" );        
        } else if (mBluetoothAdapter == null || !mBluetoothAdapter.isEnabled()) {
            BtlibPlugin.me.activity.startActivityForResult(new Intent("android.bluetooth.adapter.action.REQUEST_ENABLE"), 1);
        } else {
            doStartScan();
        }
    }

    public boolean connect(String address) {
        this.mBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
        if (this.mBluetoothAdapter == null || address == null) {
            Log.d(TAG,"connect:BluetoothAdapter not initialized or unspecified address.");
            return false;
        } else {
            BluetoothDevice device = this.mBluetoothAdapter.getRemoteDevice(address);
            if (device == null) {
                Log.e(TAG,"connect:Device not found.  Unable to connect.");
                return false;
            }
            this.mBluetoothGatt = device.connectGatt(BtlibPlugin.me.activity, false, this.mGattCallback);
            Log.d(TAG,"connect:Trying to create a new connection.");
            return true;
        }
    }

    public void disconnect(String deviceAddress) {
        if (this.mBluetoothAdapter == null || this.mBluetoothGatt == null) {
            Log.e(TAG,"disconnect:BluetoothAdapter not initialized");
        } else {
            mBluetoothGatt.disconnect();
            Log.d(TAG, "disconnect");
        }
    }

    public void readCharacteristic(BluetoothGattCharacteristic characteristic) {
        if (this.mBluetoothAdapter == null || this.mBluetoothGatt == null) {
            Log.e(TAG,"readCharacteristic:BluetoothAdapter not initialized");
        } else {
            this.mBluetoothGatt.readCharacteristic(characteristic);
        }
    }

    public void writeCharacteristic(BluetoothGattCharacteristic characteristic) {
        if (this.mBluetoothAdapter == null || this.mBluetoothGatt == null) {
            Log.e(TAG,"writeCharacteristic:BluetoothAdapter not initialized");
        } else {
            this.mBluetoothGatt.writeCharacteristic(characteristic);
        }
    }

    public void setCharacteristicNotification(BluetoothGattCharacteristic characteristic, boolean enabled) {
        if (this.mBluetoothAdapter == null || this.mBluetoothGatt == null) {
            Log.e(TAG,"setCharacteristicNotification:BluetoothAdapter not initialized");
            return;
        }
        this.mBluetoothGatt.setCharacteristicNotification(characteristic, enabled);
        final UUID CLIENT_CHARACTERISTIC_CONFIG_DESCRIPTOR_UUID = UUID
                .fromString("00002902-0000-1000-8000-00805f9b34fb");
        final BluetoothGattDescriptor descriptor = characteristic
                .getDescriptor(CLIENT_CHARACTERISTIC_CONFIG_DESCRIPTOR_UUID);
        if (descriptor != null) {
            descriptor.setValue(BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE);
            this.mBluetoothGatt.writeDescriptor(descriptor);
        }
    }

    public List<BluetoothGattService> getSupportedGattServices() {
        if (this.mBluetoothGatt == null) {
            return null;
        }
        return this.mBluetoothGatt.getServices();
    }

    public boolean readRemoteRssi(String deviceAddress) {
        if (this.mBluetoothGatt == null)
            return false;
        return mBluetoothGatt.readRemoteRssi();

    }

    private final BluetoothGattCallback mGattCallback = new BluetoothGattCallback() {
        public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
            if (newState == 2) {
                BtlibPlugin.me.ThrowEvent(
                    BTEvent.BLE_CONNECTED, 
                    "address", gatt.getDevice().getAddress()
                    );
                    Log.d("BluetoothLeService", "Connected to GATT server.");
                Log.d("BluetoothLeService",
                        "Attempting to start service discovery:" + gatt.discoverServices());
            } else if (newState == 0) {
                BtlibPlugin.me.ThrowEvent(
                    BTEvent.BLE_DISCONNECTED, "address", gatt.getDevice().getAddress()
                    );
                Log.d("BluetoothLeService", "Disconnected from GATT server.");
            }
        }

        public void onServicesDiscovered(BluetoothGatt gatt, int status) {
            if (status == 0) {
                lookupGattServices(gatt,getSupportedGattServices());
            } else {
                Log.d("BluetoothLeService", "onServicesDiscovered received: " + status);
            }
        }

        public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
            if (status == 0) {
                BtlibPlugin.me.ThrowEventBuff(
                    BTEvent.BLE_DATA_AVAILABLE, 
                    "address", gatt.getDevice().getAddress(),
                    "service", UUIDtoBTString(characteristic.getService().getUuid().toString()),
                    "characteristic", UUIDtoBTString(characteristic.getUuid().toString()),
                    "data", characteristic.getValue() 
                    );

            }
        }

        public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
            BtlibPlugin.me.ThrowEventBuff(
                BTEvent.BLE_DATA_AVAILABLE, 
                "address", gatt.getDevice().getAddress(),
                "service", UUIDtoBTString(characteristic.getService().getUuid().toString()),
                "characteristic", UUIDtoBTString(characteristic.getUuid().toString()),
                "data", characteristic.getValue()  
                 );
        }

        public void onReadRemoteRssi(BluetoothGatt gatt, int rssi, int status) {
            super.onReadRemoteRssi(gatt, rssi, status);
            Log.d("BluetoothLeService", "onReadRemoteRssi: " + rssi);
            BtlibPlugin.me.ThrowEvent(
                BTEvent.BLE_RSSI, 
                "address", gatt.getDevice().getAddress(),
                "rssi", Integer.toString(rssi)
                );
        }
    };

}
